/*************************************************************************/
/* radiative.c                                                           */
/*                                                                       */
/* Various functions helpful for radiative transfer calculations.        */
/*                                                                       */
/* Author:                                                               */
/*   Arve Kylling                                                  @au@  */
/*   NILU,                                                         @ad@  */
/*   Norway.                                                       @ad@  */
/*                                                                       */
/* ----------------------------------------------------------------------*/
/* Copyright (C) 1997 Arve Kylling                                       */
/*                                                                       */
/* This program is free software; you can redistribute it and/or modify  */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 1, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* This program is distributed in the hope that it will be useful,       */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*                                                                       */
/* To obtain a copy of the GNU General Public License write to the       */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,    */
/* USA.                                                                  */
/*-----------------------------------------------------------------------*/
/*************************************************************************/

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <malloc.h>
#include "ascii.h"
#include "radiative.h"
#include "numeric.h"

/***********************************************************************************/
/* Function: set_out                                                      @61_30i@ */
/* Description:                                                                    */
/*  Linearly interpolate the cumulative optical depth (calculated from             */
/*  @code{dtauc} to get approximate @code{utau} corresponding to altitudes         */
/*  @code{zout}.                                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float *zd:    The altitude of each level in km. @code{zd(nlyr)} is the         */
/*                bottom of the atmosphere (input).                                */
/*  float *dtauc: Optical depth of each layer (input).                             */
/*  int nlyr:     Number of atmospheric layers (input).                            */
/*  float *zout:  Array of altitudes where the output is to be calculated (input). */
/*  int nzout:    Number of output altitudes (input).                              */
/*  float *utau:  The optical depth at @code{ntau} altitudes (output).             */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int set_out (float *zd, float *dtauc, int nlyr, float *zout, int nzout, 
	     float *utau) 
{
  double* cumtau=NULL;
  int i=0, lu=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0;

  cumtau = (double *) calloc (nlyr+1, sizeof(double));
  for (lu=1; lu<=nlyr; lu++)
    cumtau[lu] = cumtau[lu-1] + (double) dtauc[lu-1]; 

  x = (double *) calloc (nlyr+1, sizeof(double));
  y = (double *) calloc (nlyr+1, sizeof(double));

  for (i=0;i<=nlyr;i++) {
    x[nlyr-i] = (double) zd[i];
    y[nlyr-i] = (double) cumtau[i];
  }

  status = linear_coeffc (x, y, nlyr+1, &a0, &a1, &a2, &a3);
  if (status!=0)  {
    fprintf (stderr, "set_out: sorry cannot do linear interpolation\n");
    fprintf (stderr, "linear_coeffc() returned status %d\n", status);
    return status;
  }

  for (i=0; i<nzout; i++)  {
    status = calc_splined_value ((double) zout[i], &ynew, x, nlyr+1, a0, a1, a2, a3);
    if (status!=0)  {
      fprintf (stderr, "set_out: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    utau[i] = (float) ynew;
    if (utau[i] < 0) utau[i] = 0; /* Optical depth should not be negative, set
				     to zero to avoid round-off problems */

  }

  /* free memory */
  free(a0); free(a1); free(a2); free(a3);
  free(x); free(y);
  free(cumtau);

  return 0;
}


