/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/sun.h"



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_extraterrestrial(char* extraterrestrial_filename, 
				 wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				 float wl_start, float wl_end);




/**************************************************************/
/* Setup extraterrestrial irradiance.                         */
/**************************************************************/

int setup_extraterrestrial (input_struct input, output_struct *output)
{
  int status=0, iv=0;
  char temp[FILENAME_MAX]="";

  /* For special applications, pre-defined extraterrestrial */
  /* spectra are used.                                      */

  /* construct location of Kato et al. [1999] extraterrestrial flux */
  strcpy (temp, input.filename[FN_PATH]);
  
  switch (input.ck_scheme) {
  case CK_KATO:
    strcat (temp, "solar_flux/kato");
    
    /* check if we need to replace the extraterrestrial flux */
    if (strcmp(temp, input.filename[FN_EXTRATERRESTRIAL])) {
      strcpy (input.filename[FN_EXTRATERRESTRIAL], temp);
      fprintf (stderr, " ... replaced extraterrestrial flux by %s\n",
	       input.filename[FN_EXTRATERRESTRIAL]);
    }
    
    break;

  case CK_FU:
    strcat (temp, "solar_flux/fu");

    /* check if we need to replace the extraterrestrial flux */
    if (strcmp(temp, input.filename[FN_EXTRATERRESTRIAL])) {
      strcpy (input.filename[FN_EXTRATERRESTRIAL], temp);
      fprintf (stderr, " ... replaced extraterrestrial flux by %s\n",
	       input.filename[FN_EXTRATERRESTRIAL]);
    }

    break;

  case CK_AVHRR_KRATZ:
    strcat (temp, "solar_flux/kratz");

    /* check if we need to replace the extraterrestrial flux */
    if (strcmp(temp, input.filename[FN_EXTRATERRESTRIAL])) {
      strcpy (input.filename[FN_EXTRATERRESTRIAL], temp);
      fprintf (stderr, " ... replaced extraterrestrial flux by %s\n",
	       input.filename[FN_EXTRATERRESTRIAL]);
    }

    break;

  case CK_NONE:
    break;

  default:
    fprintf (stderr, " ... error: unsupported correlated-k scheme\n");
    return -1;
  }


  

  /**** Read extraterrestrial irradiance ****/
  status = read_extraterrestrial(input.filename[FN_EXTRATERRESTRIAL], 
				 &input.wl, &output->wl, 
				 input.wl.wl_start, input.wl.wl_end);

  if (status!=0) {
    fprintf (stderr, " ... error %d reading %s\n",
	     status, input.filename[FN_EXTRATERRESTRIAL]);
    return status;
  }

  /**** Correction factor for Earth-Sun distance ****/
  output->eccentricity_factor = 1.0;
  
  if (input.day_of_year > 0.0) {
    output->eccentricity_factor = eccentricity (input.day_of_year);
    for (iv=0; iv<output->wl.nlambda_h; iv++)
      output->wl.fbeam[iv] *= output->eccentricity_factor;
  }

  return 0;
}



/**************************************************************/
/* Read extraterrestrial irradiance.                          */
/**************************************************************/

static int read_extraterrestrial (char* extraterrestrial_filename, 
				  wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				  float wl_start, float wl_end) 
{
  int iv=0, nlambda=0, status=0;
  float *tmp_lambda=NULL, *tmp_fbeam=NULL;
  
  /* read extraterrestrial irradiance */
  status = read_2c_file_float (extraterrestrial_filename, 
			       &tmp_lambda, &tmp_fbeam, &(wl_out->nlambda_h));
  
  if (status!=0) {
    fprintf (stderr, " ... error %d opening %s\n", status, extraterrestrial_filename);
    return status;
  }

  
  /* count the extraterrestrial wavelengths within the */
  /* user-defined wavelength range                     */
  nlambda = 0;
  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end))
      nlambda++;

  wl_out->lambda_h = (float *) calloc (nlambda, sizeof(float));
  wl_out->fbeam    = (float *) calloc (nlambda, sizeof(float));

  wl_out->map_e2h  = (int *)   calloc (nlambda, sizeof(int));

  nlambda = 0;
  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end)) {
      wl_out->lambda_h[nlambda] = tmp_lambda[iv];
      wl_out->fbeam[nlambda]    = tmp_fbeam[iv];
      wl_out->map_e2h[nlambda]  = iv;
      
      nlambda++;
    }

  wl_out->nlambda_h = nlambda;
  
  /* free memory */
  free(tmp_lambda);
  free(tmp_fbeam);

  return 0;
}



